import numpy as np
from anko import stats_util
import copy
# TODO: remove tmp comment block

class AnomalyDetector:
    
    def __init__(self, t, series):
        if isinstance(t, list): t = np.array(t)
        if isinstance(series, list): series = np.array(series)
        self.t = t
        self.t_scaleless = np.arange(1, len(series)+1)
        self.series = series
        self.clone_series = copy.deepcopy(series)
        self.using_boxcox = False
        self.using_z_normalization = False
        self.check_failed = True
        self.info_criterion = 'AIC'
        self.thres_params = {
            "p_normality": 1e-3,
            "skewness": 20,
            "normal_std_width": 3,
            "normal_std_err": 1,
            "normal_err": 1e+1,
            "max_mag": 1e+3,
            "linregress_slope": 0.1,
            "linregress_std_err": 1e+1,
            "linregress_res": 3,
            "step_func_err": 1e+1,
            "step_func_res": 3,
            "exp_decay_err": 1e+1,
            "exp_decay_res": 2,
            "linearity": 1e-2,
            "min_res": 10
        }
        self.error_code = {
                "0": "Check passed.",
                "-1": "ConvergenceError: Gaussian fitting may not converge, std_err > std_err_th.",
                "-2": "Warning: Normal distribution may have skewed, skewness > skewness_th.",
                "-3": "ConvergenceError: General erf fitting may not converge, perr > perr_th.",
                "-4": "ConvergenceError: Exponential fitting may not converge, perr > perr_th.",
                "-5": "ConvergenceError: Linear ansatz fitting may not converge, perr > perr_th.",
                "-6": "Warning: Rawdata might be oscillating, data flips sign repeatedly over mean.",
                "-7": "Info: build_statsdata is using boxcox method.",
                "-8": "Info: build_statsdata is using z normalization.",
                "-9": "Info: There are more than %d discontinuous points detected."
        }
    
    def _build_stats_model(self):
        ref = {}; IC_score = {}; lmbda = 1; proceed = False           
        histo_x, histo_y = stats_util.get_histogram(self.series)
        
        try:    
            normality = stats_util.normaltest(histo_y)
        except ValueError:
            normality = [np.inf, np.inf]     
        
        if normality[1] >= self.thres_params["p_normality"] and np.isfinite(normality[1]):   
# =============================================================================
#             # Check the magnitude of input data is not too large, else use log-boxcox transformation.
#             if np.mean(self.series) > self.thres_params["max_mag"]: 
#                 lmbda = 0; self.using_boxcox = True               
# =============================================================================
            try:
                # Perform a Gaussian fit analysis, where y(x) is the histogram built from series
                ref["model"] = 'gaussian'
                ref["popt"], ref["perr"] = stats_util.gaussian_fit(self.series, lmbda)
            except TypeError:
                # Caused by the lake of minimum samples 
                # Treat as flat histogram, appending mode as mean and np.inf as std. All perrs are set to be 0.
                if stats_util.data_is_linear(histo_x, histo_y, self.thres_params["linearity"]):
                    ref["model"] = 'flat_histo'
                    ref["popt"], ref["perr"] = stats_util.flat_histogram(self.series)
                else:
                    pass
            except RuntimeError:
                pass
        
        if "popt" in ref:
            err_score = np.sum(np.square(ref["perr"]))
            if err_score > self.thres_params["normal_err"]: 
                proceed = True
        else:
            proceed = True
        
        if proceed:
# =============================================================================
#             # Check the magnitude of input data is not too large, else perform z normalization
#             if np.mean(self.series) > self.thres_params["max_mag"]: 
#                 self.series = stats_util.z_normalization(self.series)
#                 self.using_z_normalization = True
#                 self.using_boxcox = False
# =============================================================================
            try:
                erf_popt, erf_perr = stats_util.general_erf_fit(self.t_scaleless, self.series)
                erf_y_pred = stats_util.general_erf(self.t_scaleless, *erf_popt.tolist())
            except RuntimeError:
                erf_popt = erf_perr = np.inf * np.ones(3)
                erf_y_pred = np.inf * np.ones(len(self.series))
                
# =============================================================================
#             ThrS_popt, ThrS_perr = stats_util.general_erf_fit(t, series, three_stair=True)
#             ThrS_y_pred = stats_util.three_stair_erf(t, *ThrS_popt.tolist())
# =============================================================================
            
            try:
                exp_popt, exp_perr = stats_util.exp_decay_fit(self.t_scaleless, self.series)
                exp_y_pred = stats_util.exp_decay(self.t_scaleless, *exp_popt.tolist())
            except RuntimeError:
                exp_popt = exp_perr = np.inf * np.ones(2)
                exp_y_pred = np.inf * np.ones(len(self.series))
            
            r_sq, intercept, slope, p_value, std_err = stats_util.linear_regression(self.t_scaleless, self.series)
            linregress_y_pred = np.polyval([slope,intercept], self.t_scaleless)
            
            if self.info_criterion == 'AIC':
                IC_score["step_func"] = stats_util.AIC_score(self.series, erf_y_pred, len(erf_popt))
# =============================================================================
#                 IC_score["three_stair"] = stats_util.AIC_score(series, ThrS_y_pred, len(ThrS_popt))
# =============================================================================
                IC_score["exp_decay"] = stats_util.AIC_score(self.series, exp_y_pred, len(exp_popt))
                IC_score["linear_regression"] = stats_util.AIC_score(self.series, linregress_y_pred, 2)
            elif self.info_criterion == 'BIC':
                IC_score["step_func"] = stats_util.BIC_score(self.series, erf_y_pred, len(erf_popt))
# =============================================================================
#                 IC_score["three_stair"] = stats_util.BIC_score(series, ThrS_y_pred, len(ThrS_popt))
# =============================================================================
                IC_score["exp_decay"] = stats_util.BIC_score(self.series, exp_y_pred, len(exp_popt))
                IC_score["linear_regression"] = stats_util.BIC_score(self.series, linregress_y_pred, 2)
            best_model = min(IC_score.items(), key=lambda x: x[1])
            if np.isclose(best_model[1],IC_score["linear_regression"],rtol=1e-2):
                best_model = "linear_regression"
            else:
                best_model = best_model[0]
            if best_model == 'step_func':
                if erf_popt[1]-erf_popt[0] > 0:
                    ref["model"] = "increase_step_func"
                else:
                    ref["model"] = "decrease_step_func"
            else:
                ref["model"] = best_model
            if best_model == 'step_func':
                ref["popt"], ref["perr"] = erf_popt, erf_perr
# =============================================================================
#             elif best_model == 'three_stair':
#                 ref["popt"], ref["perr"] = ThrS_popt, ThrS_perr
# =============================================================================
            elif best_model == 'exp_decay':
                ref["popt"], ref["perr"] = exp_popt, exp_perr
            elif best_model == 'linear_regression':
                ref["popt"], ref["perr"] = [intercept, slope], std_err    
        return ref
    
    def check(self):
        statsdata = self._build_stats_model()
        model_id = statsdata["model"]
        anomalous_data = "None"; msgs = []
            
        if model_id == 'gaussian' or model_id == 'flat_histo': 
            histo_x, histo_y = stats_util.get_histogram(self.clone_series)
            if statsdata["perr"][2] > self.thres_params["normal_std_err"]:
                msgs.append(self.error_code["-1"])
            if self.using_boxcox: 
                histo_x = stats_util.boxcox(histo_x, lmbda=0)
            # Large error exception: if caused by extremely flat histogram  
            if stats_util.data_is_linear(histo_x, histo_y, self.thres_params["linearity"]):
                self.check_failed = False
                return
            # Get anomalous data
            std_th = max(self.thres_params["normal_std_width"] * statsdata["popt"][2], self.thres_params["min_res"])
            allowed_domain = (statsdata["popt"][1]-std_th, statsdata["popt"][1]+std_th)
            anomalous_idx = np.where(np.logical_or(histo_x<=allowed_domain[0], histo_x>=allowed_domain[1]))[0] 
            if len(anomalous_idx) != 0: 
                anomalous_data = histo_x[anomalous_idx]
                if self.using_boxcox: anomalous_data = np.exp(anomalous_data)
                anomalous_t = self.t[np.where(np.in1d(anomalous_data, self.series))]
            if abs(stats_util.skew(histo_y)) > self.thres_params["skewness"]:
                msgs.append(self.error_code["-2"])
                    
        elif model_id == "increase_step_func":
            err_score = np.sum(np.square(statsdata["perr"]))
            if err_score > self.thres_params["step_func_err"]:
                msgs.append(self.error_code["-3"]) 
            res = stats_util.fitting_residual(self.t_scaleless, self.series, stats_util.general_erf, statsdata["popt"],
                                              mask_as_zero=True, min_res=self.thres_params["min_res"])
            anomalous_t = self.t[res > self.thres_params["step_func_res"]]
            anomalous_data = self.clone_series[res > self.thres_params["step_func_res"]]                                      
          
        elif model_id == "decrease_step_func":
            err_score = np.sum(np.square(statsdata["perr"]))
            if err_score > self.thres_params["step_func_err"]:    
                res = stats_util.fitting_residual(self.t_scaleless, self.series, stats_util.general_erf, statsdata["popt"],
                                                  mask_as_zero=True, min_res=self.thres_params["min_res"])
                anomalous_t = self.t[res > self.thres_params["step_func_res"]]
                anomalous_data = self.series[res > self.thres_params["step_func_res"]]
                msgs.append(self.error_code["-3"])
            else:   
                anomalous_idx = np.where(self.t_scaleless > statsdata["popt"][2])[0]
                if len(anomalous_idx) != 0 and (statsdata["popt"][0] - statsdata["popt"][1]) > self.thres_params["min_res"]: 
                    anomalous_t = self.t[anomalous_idx]
                    anomalous_data = self.clone_series[anomalous_idx]
            
# =============================================================================
#         elif model_id == 'three_stair':
#             err_score = np.sum(np.square(statsdata[key]["perr"]))
#             if err_score > self.thres_params["step_func_err"]:
#                 msgs.append(self.dyError.getErrorText(16))
#             t = np.arange(1, len(statsdata[key]["series"])+1) 
#             res = stats_util.fitting_residual(t, statsdata[key]["series"], stats_util.three_stair_erf, statsdata[key]["popt"])
#             anomalous_data = statsdata[key]["series"][res > self.thres_params["step_func_res"]]                  
# =============================================================================
            
        elif model_id == 'exp_decay':
            #err_score = np.sum(np.square(statsdata[key]["perr"]))
            err_score = statsdata["perr"][1]
            if err_score > self.thres_params["exp_decay_err"]:
                msgs.append(self.error_code["-4"])
            res = stats_util.fitting_residual(self.t_scaleless, self.series, stats_util.exp_decay, statsdata["popt"],
                                              mask_as_zero=True, min_res=self.thres_params["min_res"])
            anomalous_t = self.t[res > self.thres_params["exp_decay_res"]]
            anomalous_data = self.clone_series[res > self.thres_params["exp_decay_res"]]                   
             
        elif model_id == 'linear_regression':
            if statsdata["perr"] > self.thres_params["linregress_std_err"]:
                msgs.append(self.error_code["-5"])
            func = lambda x, a, b: a + b*x 
            res = stats_util.fitting_residual(self.t_scaleless, self.series, func, statsdata["popt"],
                                              mask_as_zero=True, min_res=self.thres_params["min_res"])
            anomalous_t = self.t[res > self.thres_params["linregress_res"]]
            anomalous_data = self.clone_series[res > self.thres_params["linregress_res"]]                  
                
        # Extra info
        if stats_util.is_oscillating(self.series): 
            msgs.append(self.error_code["-6"])
        if self.using_boxcox: 
            msgs.append(self.error_code["-7"])
        if self.using_z_normalization:
            msgs.append(self.error_code["-8"])    
        discontinuity = len(stats_util.discontinuous_idx(self.series))
        if discontinuity > 0:
            msgs.append(self.error_code["-9"] %discontinuity)
                
        if anomalous_data == "None" or anomalous_data.size == 0: 
            self.check_failed = False
            msgs.append(self.error_code["0"])
        
        statsdata["anomalous_data"] = list(zip(anomalous_t, anomalous_data))
        statsdata["extra_info"] = msgs
        return statsdata



import numpy as np
import collections, itertools
from scipy.stats import boxcox, linregress, skew, normaltest
from scipy.optimize import curve_fit
# TODO: handle typing for returning tuple

def get_histogram(x, sort_histo: bool=False):
    """
    Return the corresponding histogram of the data x.
    
    Args:
        x (numpy.ndarray): One-dimensional array of data.
        sort_histo (bool, optional): If True return the sorted histogram.
        
    Returns:
        keys: Set of data x (no duplicate).
        vals: Number of appearance for each key in keys. 
    """
    counter = collections.Counter(x)
    keys = np.fromiter(counter.keys(), dtype=float)
    vals = np.fromiter(counter.values(), dtype=float)
    if sort_histo:
        keys_idx = np.argsort(keys)
        keys = keys[keys_idx]
        vals = vals[keys_idx]
    return keys, vals
    
def normal_distr(x: np.ndarray, a: float, x0: float, sigma: float) -> np.ndarray:
    """
    Calculate normal distribution of input array x.
    
    Args:
        x (numpy.ndarray): Input values.
        a (float): Overall normalization constant.
        x0 (float): Mean.
        sigma (float): Standard deviation.
            
    Returns:
        out (numpy.ndarray): Output array. 
    """
    return a * np.exp(-(x-x0)**2/(2*sigma**2))

def gaussian_fit(x: np.ndarray, lmbda: float=1, sort_histo: bool=False, half: str=None, maxfev: int=2000, bounds=[0,1e+6]):
    """
    Fitting the Gaussian (normal) distribution for input data x.
    
    Args:
        x (numpy.ndarray): Input values.    
        lmbda (float, optional): If not equal to 1, \
            perform BoxCox transformation with parameter lmbda to input x. \
            This is useful to make the data more normal-distribution-like. 
        sort_histo (bool, optional): If True use the sorted histogram.
        maxfev (int, optional): Maximum step of fitting iteration.
    
    Returns:
        popt (numpy.ndarray): Estimate value of a, x0 and sigma of Gaussian distribution.
        perr (numpy.ndarray): Error of popt. Defined by the square of diagonal element of covariance matrix.
    """
    keys, vals = get_histogram(x, sort_histo)
    if lmbda != 1: keys = boxcox(keys, lmbda)
    a_sg = max(vals) * 0.2
    m_sg = np.median(x)
    std_sg = (max(keys)-min(keys))/4
    if half == 'left':
        popt, pcov = curve_fit(left_half_normal_distr,keys,vals,p0=[a_sg,m_sg,std_sg],maxfev=maxfev,bounds=bounds)
    elif half == 'right':
        popt, pcov = curve_fit(right_half_normal_distr,keys,vals,p0=[a_sg,m_sg,std_sg],maxfev=maxfev,bounds=bounds)
    else:
        popt, pcov = curve_fit(normal_distr,keys,vals,p0=[a_sg,m_sg,std_sg],maxfev=maxfev,bounds=bounds)
    perr = np.sqrt(np.diag(pcov))
    return popt, perr

def left_half_normal_distr(x: np.ndarray, a: float, x0: float, sigma: float) -> np.ndarray:
    """
    Calculate left-side half normal distribution of input array x.
    
    Args:
        x (numpy.ndarray): Input values.
        a (float): Overall normalization constant.
        x0 (float): Mean.
        sigma (float): Standard deviation.
    Returns:
        out (numpy.ndarray): Output array. 
    """
    return a * np.multiply(np.exp(-(x-x0)**2/(2*sigma**2)), -1*np.heaviside(x-x0, 0.5))

def right_half_normal_distr(x: np.ndarray, a: float, x0: float, sigma: float) -> np.ndarray:
    """
    Calculate right-side half normal distribution of input array x.
    
    Args:
        x (numpy.ndarray): Input values.
        a (float): Overall normalization constant.
        x0 (float): Mean.
        sigma (float): Standard deviation.
    Returns:
        out (numpy.ndarray): Output array. 
    """
    return a * np.multiply(np.exp(-(x-x0)**2/(2*sigma**2)), np.heaviside(x-x0, 0.5))

def flat_histogram(x: np.ndarray):
    """
    Manually assign parameters of Gaussian distrinution if the given histogram is too flat. \
    In this senario the histogram of data is regarded as a local segment of a larger normal-distribution-like histogram, \
    with standard deviation which exceeds the current consideration of domain. 
    
    Parameters of Gaussian distribution are assigned as following:
        1. Number of appearance of mode as normalization constant, a. 
        2. Mode of data x as mean, x0. 
        3. Standard deviation is set to infinity (numpy.inf).
    
    Args:
        x (numpy.ndarray): Input values.
    Returns:
        popt (numpy.ndarray): Assigned values for Gaussian distribution. 
        perr (numpy.ndarray): Errors are set to zero.
    """
    counter = collections.Counter(x)
    mode = counter.most_common()
    popt = np.array([mode[0][1], mode[0][0], np.inf])
    perr = np.array([0, 0, 0])
    return popt, perr

def linear_regression(x: np.ndarray, y: np.ndarray, lmbda: float=1):
    """
    Fitting linear ansatz for input data (x, y). 
    
    y = intercept + slope * x.
    
    Args:
        x (numpy.ndarray): x coordinate of input data points. 
        y (numpy.ndarray): y coordinate of input data points.
        lmbda (float, optional): If not equal to 1, \
            perform BoxCox transformation with parameter lmbda to input y. \
            This is useful to lower and rescale the magnitude of data. 
    Returns:
        r_sq (float): Coefficient of determination.
        intercept (float): Intercept of the regression line.
        slope (float): Slope of the regression line.
        p_value (float): Two-sided p-value for a hypothesis test whose null hypothesis is that the slope is zero, \
            using Wald Test with t-distribution of the test statistic.
        std_err (float): Standard error of the estimated gradient.
    """
    if lmbda != 1: y = boxcox(y, lmbda)
    slope, intercept, r_value, p_value, std_err = linregress(x, y)
    r_sq = r_value**2
    return r_sq, intercept, slope, p_value, std_err

def data_is_linear(x: np.ndarray, y: np.ndarray, std_err_th: float=1e-2) -> bool:
    """
    Check whether the data (x, y) is linear under the given tolerance. \
    This will perform a linear regression fitting.
    
    Args:
        x (numpy.ndarray): x coordinate of input data points. 
        y (numpy.ndarray): y coordinate of input data points.
        std_err_th (float, optional): Threshold value of std_err.
    Returns:
        out (bool): Return Ture if data is flat, else return False.
    """
    r_sq, intercept, slope, p_value, std_err = linear_regression(x, y)
    if p_value == 1 or std_err < std_err_th:
        return True
    else:
        return False

def general_erf(x: np.ndarray, a: float, b: float, x0: float) -> np.ndarray:
    """
    Calculate the generalize error function of input array x.
    
    f(x) = a; x < x0,
           (a+b)/2; x = x0,
           b; x > x0
    
    Args:
        x (numpy.ndarray): Input values.
        a (float): Value of first stair. 
        b (float): Value of second stair.
        x0 (float): Location of the cliff.
    Returns:
        out (numpy.ndarray): Output array.
    """
    return (abs(b-a)/2) * np.sign(x-x0) + (a+b)/2

def three_stair_erf(x, c0, c1, c2, x1, x2):
    return c0*np.sign(x-x1) + c1*np.sign(x-x2) + c2
    
def general_erf_fit(x, y, lmbda=1, three_stair=False, maxfev=2000, bounds=[0,1e+6]):
    """
    Fitting generalize error function for input data (x, y).
    
    Args:
        x (numpy.ndarray):
        y (numpy.ndarray):
        lmbda (float):
        three_stair (bool):
        maxfev (int):
        bounds (list[float]):
    Returns:
        popt (numpy.ndarray):
        perr (numpy.ndarray):
    """
    if lmbda != 1: y = boxcox(y, lmbda)
    if three_stair:
        a_sg = y[0]; b_sg = y[int(len(y)/2)]; c_sg = y[-1]
        c0_sg = (b_sg-a_sg)/2; c1_sg = (c_sg-b_sg)/2; c2_sg = (a_sg + c_sg)/2
        x1_sg, x2_sg = np.sort(np.diff(y))[::-1][:2]     
        popt, pcov = curve_fit(three_stair_erf,x,y,p0=[c0_sg,c1_sg,c2_sg,x1_sg,x2_sg],maxfev=maxfev,bounds=bounds)
        perr = np.sqrt(np.diag(pcov))
    else:
        a_sg = y[0]; b_sg = y[-1]; x0_sg = x[np.argmax(np.diff(y))]
        popt, pcov = curve_fit(general_erf,x,y,p0=[a_sg,b_sg,x0_sg],maxfev=maxfev,bounds=bounds)
        perr = np.sqrt(np.diag(pcov))
    return popt, perr    

def exp_decay(x, a, alpha):
    # Domain: x >= 0
    if len(np.where(np.array(x) < 0)[0]) != 0:
        raise ValueError("Domain of exp(-x) is restricted in x > 0.")
    return a * np.exp(-alpha*x)
    
def exp_decay_fit(x, y, lmbda=1, maxfev=2000):
    if lmbda != 1: y = boxcox(y, lmbda)
    popt, pcov = curve_fit(exp_decay,x,y,maxfev=maxfev)
    perr = np.sqrt(np.diag(pcov))
    return popt, perr    

def smoothness(x, normalize=False):
    # TODO: this definition is not good
    dx = np.diff(x)
    sm = np.std(dx)
    if normalize: sm /= abs(np.mean(dx))
    return sm
    
def discontinuous_idx(x, std_width=1):
    dx = np.diff(x)
    idx = np.where(abs(dx-np.mean(dx)) > std_width*np.std(x))
    return idx[0]

def is_oscillating(x, osci_freq_th=0.3):
    mu = np.mean(x)
    y = x - mu
    ocsi_times = len(list(itertools.groupby(y, lambda y: y>0)))
    ocsi_freq = ocsi_times/len(x)
    if ocsi_freq > osci_freq_th:
        return True
    else:
        return False
   
def fitting_residual(x, y, func, args, standardized=True, mask_as_zero=False, min_res=10):
    y_predict = func(x, *args)
    res = np.subtract(y, y_predict)
    norm = np.std(res)
    if mask_as_zero:
        res[np.where(abs(res) < min_res)] = 0
    if standardized and norm != 0:
        res /= norm
    return abs(res)
    
def AIC_score(y, y_predict, p):
    n = len(y)
    res = np.subtract(y, y_predict)
    rss = np.sum(np.power(res, 2))
    aic_score = n*np.log(rss/n) + 2*p
    return aic_score    

def BIC_score(y, y_predict, p):
    n = len(y)
    res = np.subtract(y, y_predict)
    rss = np.sum(np.power(res, 2))
    bic_score = n*np.log(rss/n) + p*np.log(n)
    return bic_score   

def z_normalization(x):
    return (x-np.mean(x))/np.std(x)




from distutils.core import setup
ver = '0.0.6'

setup(
  name = 'anko',
  packages = ['anko'],
  version = ver,
  license='MIT',
  description = 'Toolkit for performing anomaly detection algorithm on time series.',
  author = 'tao-lin',
  author_email = 'tanlin2013@gmail.com', 
  url = 'https://github.com/tanlin2013/anko',
  download_url = 'https://github.com/tanlin2013/anko/archive/v%s.tar.gz' %ver,
  keywords = ['statistics', 'time series', 'anomaly detection'],
  install_requires=[
          'numpy==1.16.4',
          'scipy==1.2.1',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Scientific/Engineering :: Mathematics',
    'License :: OSI Approved :: MIT License',   # Pick a license
    'Programming Language :: Python :: 3',      # Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)

